/* main.c -- send a reset in response to TCP SYN packets from Netlink.
 *
 * Copyright (c) 2000 Bellamy Consulting Limited
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: main.c,v 1.9 2000/08/14 05:06:55 planet Exp $
 *
 * Changes:
 *   
 *   2000-08-10
 *
 *   Firewall mark code added by Patrick Boyne <pjb169@psu.edu>, via code
 *   borrowed from fwdumpd (http://indev.insu.com/Fwctl/fwdumpd.html).
 *   (Modified by NDB; allow mark to be specified in decimal, octal or hex)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#define VERSION			"1.1"
#define DEFAULT_NETLINK_DEVICE	"/dev/netlink"
#define DPRINTF(args...)	{ if (g_debug) fprintf(stderr, ##args); }
#define SYSLOG(args...)	{ if (g_use_syslog) syslog(LOG_WARNING, ##args); }

struct netlinkhdr {
	unsigned int len;
	unsigned int mark;
	char iface[IFNAMSIZ];
};

static int g_debug = 0;
static int g_use_fwmark = 0;
static int g_use_syslog = 0;
static int g_log_packets = 0;
static u_int16_t g_ip_id;
static u_int32_t g_fwmark;

static void usage(const char *);
static void version(const char *);
static int main_loop(int, int);
static void send_reset(int, struct iphdr *, struct tcphdr *);
static int ipv4_checksum_add(const void *, size_t);
static u_int16_t ipv4_checksum_final(int);
static int get_mark_value(const char *);
static char *ip_to_char(u_int32_t);

int main(int argc, char *argv[])
{
	int r;
	int fd, rawsock;
	int daemon_mode;
	char on;
	char *netlink_device = DEFAULT_NETLINK_DEVICE;
	char *pid_file = NULL;
	FILE *fp;

	daemon_mode = 1;
	srand(time(NULL));
	g_ip_id = rand() & 0xffff;

	while ((r = getopt(argc, argv, "hvdsf:p:m:")) != EOF) {
		switch (r) {
			case 'h':
				usage(argv[0]);
				exit(1);
			case 'v':
				version(argv[0]);
				exit(1);
			case 'd':
				g_debug = 1;
				DPRINTF("Debugging enabled.\n");
				break;
			case 's':
				g_log_packets = 1;
				break;
			case 'f':
				netlink_device = optarg;
				break;
			case 'p':
				pid_file = optarg;
				break;
			case 'm':
				if (get_mark_value(optarg) != 0) {
					fprintf(stderr, "Invalid mark "
							"value.\n");
					exit(1);
				}
				g_use_fwmark = 1;
				break;
		}
	}

	DPRINTF("%s built at %s on %s\n", argv[0], __TIME__, __DATE__);

	DPRINTF("Using netlink device: %s\n", netlink_device);

	if (g_use_fwmark)
		DPRINTF("Only matching packets with mark 0x%.8x\n", g_fwmark);

	fd = open(netlink_device, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "Failed to open %s: %s\n",
				netlink_device, strerror(errno));
		exit(1);
	}
	DPRINTF("Netlink device opened OK.\n");

	rawsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (rawsock < 0) {
		fprintf(stderr, "Failed to open raw socket: %s\n",
				strerror(errno));
		exit(1);
	}
	DPRINTF("Raw socket opened OK.\n");

	on = 1;
	if (setsockopt(rawsock, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on))) {
		fprintf(stderr, "Failed to set IP_HDRINCL on raw socket: %s\n",
				strerror(errno));
		exit(1);
	}
	DPRINTF("Raw socket set to IP_HDRINCL OK.\n");

	if (g_debug == 0) {
		daemon(0, 0);
		g_use_syslog = 1;
		openlog("return-rst", LOG_PID, LOG_DAEMON);
		if (pid_file != NULL) {
			fp = fopen(pid_file, "w+");
			if (fp != NULL) {
				fprintf(fp, "%d\n", getpid());
				fclose(fp);
			}
		}
	}

	r = main_loop(fd, rawsock);

	close(fd);
	close(rawsock);
	
	exit(r);
}

static int main_loop(int nlfd, int rs)
{
	int b;
	char buf[BUFSIZ];
	char *src, *dst;
	struct netlinkhdr *nlh = (struct netlinkhdr *) buf;
	struct iphdr *iph = (struct iphdr *) (buf + sizeof(struct netlinkhdr));
	struct tcphdr *tcph;

	while ((b = read(nlfd, buf, sizeof(buf))) > 0) {
		/* If -m is specified, check the netlink marker, and discard
		 * packets that do not match.
		 */
		if (g_use_fwmark && nlh->mark != g_fwmark) {
			DPRINTF("Discarding packet, mark 0x%.8x != 0x%.8x\n",
					nlh->mark, g_fwmark);
			continue;
		}

		/* Check the packet is big enough to contain IP + TCP data */
		if (b < sizeof(struct netlinkhdr)
				+ (iph->ihl << 2)
				+ sizeof(struct tcphdr)) {
			DPRINTF("Short packet, ignoring.\n");
			continue;
		}
		DPRINTF("Netlink: iface %s, len %d, mark 0x%.8x\n",
				nlh->iface, nlh->len, nlh->mark);

		if (iph->protocol != IPPROTO_TCP) {
			DPRINTF("Packet: not TCP\n");
			continue;
		}
		
		tcph = (struct tcphdr *) (buf + sizeof(struct netlinkhdr)
				+ (iph->ihl << 2));
		
		if (!tcph->syn) {
			DPRINTF("Packet: no SYN\n");
			continue;
		}
		if (tcph->ack) {
			DPRINTF("Packet: ACK set\n");
			continue;
		}
		DPRINTF("src = 0x%.8x\n", ntohl(iph->saddr));
		DPRINTF("dst = 0x%.8x\n", ntohl(iph->daddr));
		DPRINTF("sport = %d\n", ntohs(tcph->source));
		DPRINTF("dport = %d\n", ntohs(tcph->dest));
		send_reset(rs, iph, tcph);
		if (g_log_packets) {
			src = ip_to_char(iph->daddr);
			dst = ip_to_char(iph->saddr);
			syslog(LOG_NOTICE, "%s: %s:%d -> %s:%d blocked with RST\n",
					nlh->iface, src, ntohs(tcph->dest),
					dst, ntohs(tcph->source));
			free(src);
			free(dst);
		}
	}
	if (b < 0)
		return 1;

	return 0;
}

static void send_reset(int rs, struct iphdr *iph, struct tcphdr *tcph)
{
	int r, sum;
	size_t len;
	u_int16_t tport;
	u_int32_t taddr;
	struct sockaddr_in sa;

	len = (iph->ihl << 2) + sizeof(struct tcphdr);

	/* Swap src and dst in IP header */
	taddr = iph->saddr;
	iph->saddr = iph->daddr;
	iph->daddr = taddr;
	
	iph->ttl = 255;
	iph->frag_off = htons(ntohs(iph->frag_off) & 0x1fff) | htons(IP_DF);
	iph->id = htons(g_ip_id);

	/* Increment the IP identifier in a really dumb way */
	g_ip_id += rand() & 0xf;

	/* Calculate IP header checksum */
	iph->check = 0;
	sum = ipv4_checksum_add(iph, iph->ihl << 2);
	iph->check = ipv4_checksum_final(sum);

	DPRINTF("IP checksum: 0x%.4x\n", iph->check);

	/* ACK + RST */
	tcph->syn = 0;
	tcph->ack = 1;
	tcph->rst = 1;
	tcph->urg = 0;
	tcph->ack_seq = htonl(ntohl(tcph->seq) + 1);
	tcph->seq = 0;
	tcph->window = 0;

	/* Swap src and dst ports in TCP header */
	tport = tcph->source;
	tcph->source = tcph->dest;
	tcph->dest = tport;
	tcph->doff = sizeof(struct tcphdr) >> 2;

	/* Calculate TCP header checksum */
	tcph->check = 0;
	sum = ipv4_checksum_add(&iph->saddr, 8);
	sum += htons(IPPROTO_TCP + sizeof(struct tcphdr));
	sum += ipv4_checksum_add(tcph, sizeof(struct tcphdr));
	tcph->check = ipv4_checksum_final(sum);

	DPRINTF("TCP checksum: 0x%.4x\n", tcph->check);

	sa.sin_family = AF_INET;
	sa.sin_port = tcph->dest;
	sa.sin_addr.s_addr = iph->daddr;
	
	DPRINTF("Sending reset to 0x%.8x:%.4x\n",
			ntohl(iph->daddr), ntohs(tcph->dest));
	
	r = sendto(rs, iph, len, 0, (struct sockaddr *) &sa, sizeof(sa));
	
	if (r < 0) {
		DPRINTF("sendto: %s\n", strerror(errno));
		SYSLOG("sendto: %s\n", strerror(errno));
		return;
	}
	if (r != len) {
		DPRINTF("sendto: bad write (%d != %d)\n", r, len);
		SYSLOG("sendto: bad write (%d != %d)\n", r, len);
		return;
	}

	DPRINTF("Reset sent OK.\n");
}

static int ipv4_checksum_add(const void *data, size_t len)
{
	int l, s;
	u_int16_t a;
	const u_int16_t *p;

	s = 0;
	a = 0;
	l = len;
	p = data;

	while (l > 1) {
		s += *p++;
		l -= 2;
	}

	if (l == 1) {
		*(u_int8_t *)(&a) = *(u_int8_t *) p;
		s += a;
	}

	return s;
}

static u_int16_t ipv4_checksum_final(int s)
{
	s = (s >> 16) + (s & 0xffff);
	s += (s >> 16);

	return (u_int16_t) ~s;
}

static int get_mark_value(const char *string)
{
	long int r;
	char *tail;

	/* base is 0 so the mark can be specified in decimal, octal or hex */
	errno = 0;
	r = strtol(string, &tail, 0);
	if (*tail != '\0') {
		return 1;
	}
	if (errno == ERANGE || r == LONG_MIN || r == LONG_MAX) {
		return 1;
	}
	if (r > UINT_MAX || r < 0) {
		return 1;
	}
	g_fwmark = (u_int32_t) r;

	return 0;
}

static char *ip_to_char(u_int32_t ip)
{
	char *str;

	str = (char *) malloc(16);
	if (str == NULL)
		return NULL;

	ip = ntohl(ip);

	snprintf(str, 16, "%d.%d.%d.%d",
			(ip >> 24) & 0xff,
			(ip >> 16) & 0xff,
			(ip >> 8) & 0xff,
			ip & 0xff);

	return str;
}

static void usage(const char *name)
{
	char *tmp;

	if ((tmp = strrchr(name, '/')) != NULL)
		name = tmp + 1;

	fprintf(stderr, "Usage: %s [-hvds] [-f device] [-m mark] [-p pidfile]\n"
			"\n"
			"\t-h\t\tThis help.\n"
			"\t-v\t\tShow version.\n"
			"\t-d\t\tTurn on debugging mode.\n"
			"\t-s\t\tSyslog blocked packets.\n"
			"\t-f device\tUse specifed device instead of %s\n"
			"\t-m mark\t\tOnly match packets with this marker.\n"
			"\t-p pidfile\tWrite the daemon PID to this file.\n",
			name, DEFAULT_NETLINK_DEVICE);
}

static void version(const char *name)
{
	char *tmp;

	if ((tmp = strrchr(name, '/')) != NULL)
		name = tmp + 1;

	fprintf(stderr, "%s version %s\n"
			"Copyright (C) 2000 Bellamy Consulting Limited - "
			"http://www.bellamy.co.nz/\n"
			"Distributed under the terms of the "
			"GNU General Public License V2\n",
			name, VERSION);
}

/* EOF */
